@@ -155,7 +155,9 @@ static void gfx_ctx_x_vk_swap_buffers(void *data)
    {
       x->vk.context.has_acquired_swapchain = false;
       if (x->vk.swapchain == VK_NULL_HANDLE)
+      {
          retro_sleep(10);
+      }
       else
          vulkan_present(&x->vk, x->vk.context.current_swapchain_index);
    }
@@ -464,7 +466,7 @@ static void gfx_ctx_x_vk_input_driver(void *data,
 
    if (string_is_equal(input_driver, "udev"))
    {
-      *input_data = input_udev.init(joypad_name);
+      *input_data = input_driver_init_wrap(&input_udev, joypad_name);
       if (*input_data)
       {
          *input = &input_udev;
@@ -473,7 +475,7 @@ static void gfx_ctx_x_vk_input_driver(void *data,
    }
 #endif
 
-   x_input      = input_x.init(joypad_name);
+   x_input      = input_driver_init_wrap(&input_x, joypad_name);
    *input       = x_input ? &input_x : NULL;
    *input_data  = x_input;
 }
