@@ -50,6 +50,8 @@
 #define MODETEST_VAL    0xffffff00
 #endif
 
+/* TODO/FIXME - reorder this according to CODING-GUIDELINES
+ * and make sure LUT table below conforms */
 struct magic_entry
 {
    int32_t offset;
@@ -213,10 +215,10 @@ static int detect_ps1_game_sub(intfstream_t *fp,
    *game_id++ = toupper(*tmp++);
    *game_id++ = '-';
 
-   if (!isalnum(*tmp))
+   if (!ISALNUM(*tmp))
       tmp++;
 
-   while (isalnum(*tmp))
+   while (ISALNUM(*tmp))
    {
       *game_id++ = *tmp++;
       if (*tmp == '.')
@@ -491,8 +493,9 @@ int cue_find_track(const char *cue_path, bool first,
 {
    int rv;
    intfstream_info_t info;
-   char *tmp_token            = (char*)malloc(MAX_TOKEN_LEN);
-   char *last_file            = (char*)malloc(PATH_MAX_LENGTH + 1);
+   char tmp_token[MAX_TOKEN_LEN];
+   char last_file[PATH_MAX_LENGTH];
+   char cue_dir[PATH_MAX_LENGTH];
    intfstream_t *fd           = NULL;
    int64_t last_index         = -1;
    int64_t cand_index         = -1;
@@ -501,10 +504,9 @@ int cue_find_track(const char *cue_path, bool first,
    uint64_t largest             = 0;
    int64_t volatile file_size = -1;
    bool is_data               = false;
-   char *cue_dir              = (char*)malloc(PATH_MAX_LENGTH);
-   cue_dir[0]                 = '\0';
+   cue_dir[0] = last_file[0]  = '\0';
 
-   fill_pathname_basedir(cue_dir, cue_path, PATH_MAX_LENGTH);
+   fill_pathname_basedir(cue_dir, cue_path, sizeof(cue_dir));
 
    info.type                  = INTFSTREAM_FILE;
    fd                         = (intfstream_t*)intfstream_init(&info);
@@ -526,7 +528,7 @@ int cue_find_track(const char *cue_path, bool first,
 
    rv = -EINVAL;
 
-   while (get_token(fd, tmp_token, MAX_TOKEN_LEN) > 0)
+   while (get_token(fd, tmp_token, sizeof(tmp_token)) > 0)
    {
       if (string_is_equal_noncase(tmp_token, "FILE"))
       {
@@ -544,27 +546,27 @@ int cue_find_track(const char *cue_path, bool first,
                goto clean;
          }
 
-         get_token(fd, tmp_token, MAX_TOKEN_LEN);
+         get_token(fd, tmp_token, sizeof(tmp_token));
          fill_pathname_join(last_file, cue_dir,
-               tmp_token, PATH_MAX_LENGTH);
+               tmp_token, sizeof(last_file));
 
          file_size = intfstream_get_file_size(last_file);
 
-         get_token(fd, tmp_token, MAX_TOKEN_LEN);
+         get_token(fd, tmp_token, sizeof(tmp_token));
 
       }
       else if (string_is_equal_noncase(tmp_token, "TRACK"))
       {
-         get_token(fd, tmp_token, MAX_TOKEN_LEN);
-         get_token(fd, tmp_token, MAX_TOKEN_LEN);
+         get_token(fd, tmp_token, sizeof(tmp_token));
+         get_token(fd, tmp_token, sizeof(tmp_token));
          is_data = !string_is_equal_noncase(tmp_token, "AUDIO");
          ++track;
       }
       else if (string_is_equal_noncase(tmp_token, "INDEX"))
       {
          int m, s, f;
-         get_token(fd, tmp_token, MAX_TOKEN_LEN);
-         get_token(fd, tmp_token, MAX_TOKEN_LEN);
+         get_token(fd, tmp_token, sizeof(tmp_token));
+         get_token(fd, tmp_token, sizeof(tmp_token));
 
          if (sscanf(tmp_token, "%02d:%02d:%02d", &m, &s, &f) < 3)
          {
@@ -605,17 +607,11 @@ int cue_find_track(const char *cue_path, bool first,
       rv = 0;
 
 clean:
-   free(cue_dir);
-   free(tmp_token);
-   free(last_file);
    intfstream_close(fd);
    free(fd);
    return rv;
 
 error:
-   free(cue_dir);
-   free(tmp_token);
-   free(last_file);
    if (fd)
    {
       intfstream_close(fd);
@@ -627,28 +623,26 @@ error:
 bool cue_next_file(intfstream_t *fd,
       const char *cue_path, char *path, uint64_t max_len)
 {
+   char tmp_token[MAX_TOKEN_LEN];
+   char cue_dir[PATH_MAX_LENGTH];
    bool rv                    = false;
-   char *tmp_token            = (char*)malloc(MAX_TOKEN_LEN);
-   char *cue_dir              = (char*)malloc(PATH_MAX_LENGTH);
    cue_dir[0]                 = '\0';
 
-   fill_pathname_basedir(cue_dir, cue_path, PATH_MAX_LENGTH);
+   fill_pathname_basedir(cue_dir, cue_path, sizeof(cue_dir));
 
    tmp_token[0] = '\0';
 
-   while (get_token(fd, tmp_token, MAX_TOKEN_LEN) > 0)
+   while (get_token(fd, tmp_token, sizeof(tmp_token)) > 0)
    {
       if (string_is_equal_noncase(tmp_token, "FILE"))
       {
-         get_token(fd, tmp_token, MAX_TOKEN_LEN);
+         get_token(fd, tmp_token, sizeof(tmp_token));
          fill_pathname_join(path, cue_dir, tmp_token, (size_t)max_len);
          rv = true;
          break;
       }
    }
 
-   free(cue_dir);
-   free(tmp_token);
    return rv;
 }
 
@@ -657,7 +651,7 @@ int gdi_find_track(const char *gdi_path, bool first,
 {
    int rv;
    intfstream_info_t info;
-   char *tmp_token   = (char*)malloc(MAX_TOKEN_LEN);
+   char tmp_token[MAX_TOKEN_LEN];
    intfstream_t *fd  = NULL;
    uint64_t largest  = 0;
    int size          = -1;
@@ -686,20 +680,20 @@ int gdi_find_track(const char *gdi_path, bool first,
    rv = -EINVAL;
 
    /* Skip track count */
-   get_token(fd, tmp_token, MAX_TOKEN_LEN);
+   get_token(fd, tmp_token, sizeof(tmp_token));
 
    /* Track number */
-   while (get_token(fd, tmp_token, MAX_TOKEN_LEN) > 0)
+   while (get_token(fd, tmp_token, sizeof(tmp_token)) > 0)
    {
       /* Offset */
-      if (get_token(fd, tmp_token, MAX_TOKEN_LEN) <= 0)
+      if (get_token(fd, tmp_token, sizeof(tmp_token)) <= 0)
       {
          errno = EINVAL;
          goto error;
       }
 
       /* Mode */
-      if (get_token(fd, tmp_token, MAX_TOKEN_LEN) <= 0)
+      if (get_token(fd, tmp_token, sizeof(tmp_token)) <= 0)
       {
          errno = EINVAL;
          goto error;
@@ -708,7 +702,7 @@ int gdi_find_track(const char *gdi_path, bool first,
       mode = atoi(tmp_token);
 
       /* Sector size */
-      if (get_token(fd, tmp_token, MAX_TOKEN_LEN) <= 0)
+      if (get_token(fd, tmp_token, sizeof(tmp_token)) <= 0)
       {
          errno = EINVAL;
          goto error;
@@ -717,7 +711,7 @@ int gdi_find_track(const char *gdi_path, bool first,
       size = atoi(tmp_token);
 
       /* File name */
-      if (get_token(fd, tmp_token, MAX_TOKEN_LEN) <= 0)
+      if (get_token(fd, tmp_token, sizeof(tmp_token)) <= 0)
       {
          errno = EINVAL;
          goto error;
@@ -726,22 +720,19 @@ int gdi_find_track(const char *gdi_path, bool first,
       /* Check for data track */
       if (!(mode == 0 && size == 2352))
       {
-         char *last_file   = (char*)malloc(PATH_MAX_LENGTH + 1);
-         char *gdi_dir     = (char*)malloc(PATH_MAX_LENGTH);
+         char last_file[PATH_MAX_LENGTH];
+         char gdi_dir[PATH_MAX_LENGTH];
 
-         gdi_dir[0]        = '\0';
+         gdi_dir[0]        = last_file[0] = '\0';
 
-         fill_pathname_basedir(gdi_dir, gdi_path, PATH_MAX_LENGTH);
+         fill_pathname_basedir(gdi_dir, gdi_path, sizeof(gdi_dir));
 
          fill_pathname_join(last_file,
-               gdi_dir, tmp_token, PATH_MAX_LENGTH);
+               gdi_dir, tmp_token, sizeof(last_file));
          file_size = intfstream_get_file_size(last_file);
+
          if (file_size < 0)
-         {
-            free(gdi_dir);
-            free(last_file);
             goto error;
-         }
 
          if ((uint64_t)file_size > largest)
          {
@@ -751,18 +742,12 @@ int gdi_find_track(const char *gdi_path, bool first,
             largest = file_size;
 
             if (first)
-            {
-               free(gdi_dir);
-               free(last_file);
                goto clean;
-            }
          }
-         free(gdi_dir);
-         free(last_file);
       }
 
       /* Disc offset (not used?) */
-      if (get_token(fd, tmp_token, MAX_TOKEN_LEN) <= 0)
+      if (get_token(fd, tmp_token, sizeof(tmp_token)) <= 0)
       {
          errno = EINVAL;
          goto error;
@@ -770,13 +755,11 @@ int gdi_find_track(const char *gdi_path, bool first,
    }
 
 clean:
-   free(tmp_token);
    intfstream_close(fd);
    free(fd);
    return rv;
 
 error:
-   free(tmp_token);
    if (fd)
    {
       intfstream_close(fd);
@@ -788,45 +771,36 @@ error:
 bool gdi_next_file(intfstream_t *fd, const char *gdi_path,
       char *path, uint64_t max_len)
 {
+   char tmp_token[MAX_TOKEN_LEN];
    bool rv         = false;
-   char *tmp_token = (char*)malloc(MAX_TOKEN_LEN);
-   int64_t offset  = -1;
 
    tmp_token[0]    = '\0';
 
    /* Skip initial track count */
-   offset = intfstream_tell(fd);
-   if (offset == 0)
-      get_token(fd, tmp_token, MAX_TOKEN_LEN);
+   if (intfstream_tell(fd) == 0)
+      get_token(fd, tmp_token, sizeof(tmp_token));
 
-   /* Track number */
-   get_token(fd, tmp_token, MAX_TOKEN_LEN);
-   /* Offset */
-   get_token(fd, tmp_token, MAX_TOKEN_LEN);
-   /* Mode */
-   get_token(fd, tmp_token, MAX_TOKEN_LEN);
-   /* Sector size */
-   get_token(fd, tmp_token, MAX_TOKEN_LEN);
+   get_token(fd, tmp_token, sizeof(tmp_token)); /* Track number */
+   get_token(fd, tmp_token, sizeof(tmp_token)); /* Offset       */
+   get_token(fd, tmp_token, sizeof(tmp_token)); /* Mode         */
+   get_token(fd, tmp_token, sizeof(tmp_token)); /* Sector size  */
 
    /* File name */
-   if (get_token(fd, tmp_token, MAX_TOKEN_LEN) > 0)
+   if (get_token(fd, tmp_token, sizeof(tmp_token)) > 0)
    {
-      char *gdi_dir   = (char*)malloc(PATH_MAX_LENGTH);
+      char gdi_dir[PATH_MAX_LENGTH];
 
       gdi_dir[0]      = '\0';
 
-      fill_pathname_basedir(gdi_dir, gdi_path, PATH_MAX_LENGTH);
+      fill_pathname_basedir(gdi_dir, gdi_path, sizeof(gdi_dir));
 
       fill_pathname_join(path, gdi_dir, tmp_token, (size_t)max_len);
 
       rv              = true;
 
       /* Disc offset */
-      get_token(fd, tmp_token, MAX_TOKEN_LEN);
-
-      free(gdi_dir);
+      get_token(fd, tmp_token, sizeof(tmp_token));
    }
 
-   free(tmp_token);
    return rv;
 }
